@@ -3,18 +3,21 @@ import asyncio
 import errno
 import select
 import time
+from dataclasses import dataclass
 from enum import IntEnum, IntFlag
 from functools import partial
-from typing import Any, Callable, List, NewType, Optional, Tuple, Union, overload
+from typing import (Any, Callable, List, MutableMapping, NewType, Optional, Set, Tuple, Union,
+                    overload)
+from weakref import WeakValueDictionary
 
 from ._ffi import alsa, ffi
 from .address import Address, AddressType
-from .event import MIDI_BYTES_EVENTS, Event, EventType, MidiBytesEvent, _snd_seq_event_t
+from .event import MIDI_BYTES_EVENTS, Event, EventType, MidiBytesEvent, RealTime, _snd_seq_event_t
 from .exceptions import StateError
 from .port import (DEFAULT_PORT_TYPE, READ_PORT_PREFERRED_TYPES, RW_PORT, RW_PORT_PREFERRED_TYPES,
                    WRITE_PORT_PREFERRED_TYPES, Port, PortCaps, PortInfo, PortType,
                    get_port_info_sort_key)
-from .queue import Queue
+from .queue import Queue, QueueInfo, QueueStatus
 from .util import _check_alsa_error
 
 _snd_seq_t = NewType("_snd_seq_t", object)
@@ -41,9 +44,18 @@ class ClientType(IntEnum):
     KERNEL = alsa.SND_SEQ_KERNEL_CLIENT
 
 
+class SequencerType(IntEnum):
+    """Sequencer type constants."""
+    _UNSET = 0
+    HW = alsa.SND_SEQ_TYPE_HW
+    SHM = alsa.SND_SEQ_TYPE_SHM
+    INET = alsa.SND_SEQ_TYPE_INET
+
+
 _snd_seq_client_info_t = NewType("_snd_seq_client_info_t", object)
 
 
+@dataclass
 class ClientInfo:
     """Client information.
 
@@ -70,34 +82,15 @@ class ClientInfo:
     :ivar event_lost: number of lost events
     """
     client_id: int
-    name: str
-    broadcast_filter: bool
-    error_bounce: bool
-    type: Optional[ClientType]
-    card_id: Optional[int]
-    pid: Optional[int]
-    num_ports: int
-    event_lost: int
-
-    def __init__(self,
-                 client_id: int,
-                 name: str,
-                 broadcast_filter: bool = False,
-                 error_bounce: bool = False,
-                 type: ClientType = None,
-                 card_id: Optional[int] = None,
-                 pid: Optional[int] = None,
-                 num_ports: int = 0,
-                 event_lost: int = 0):
-        self.client_id = client_id
-        self.name = name
-        self.broadcast_filter = broadcast_filter
-        self.error_bounce = error_bounce
-        self.type = type
-        self.card_id = card_id
-        self.pid = pid
-        self.num_ports = num_ports
-        self.event_lost = event_lost
+    name: str = ""
+    broadcast_filter: bool = False
+    error_bounce: bool = False
+    type: Optional[ClientType] = None
+    card_id: Optional[int] = None
+    pid: Optional[int] = None
+    num_ports: int = 0
+    event_lost: int = 0
+    event_filter: Optional[Set[EventType]] = None
 
     @classmethod
     def _from_alsa(cls, info: _snd_seq_client_info_t):
@@ -117,6 +110,18 @@ class ClientInfo:
         except AttributeError:
             pid = None
         name = ffi.string(alsa.snd_seq_client_info_get_name(info))
+
+        # deprecated but seems the only way to check if there is a filter
+        has_filter = bool(alsa.snd_seq_client_info_get_event_filter(info))
+        if has_filter:
+            event_filter = set()
+            for e_type in EventType:
+                if e_type == EventType.NONE:
+                    continue
+                if alsa.snd_seq_client_info_event_filter_check(info, e_type):
+                    event_filter.add(e_type)
+        else:
+            event_filter = None
         return cls(
                 client_id=alsa.snd_seq_client_info_get_client(info),
                 name=name.decode(),
@@ -127,6 +132,7 @@ class ClientInfo:
                 pid=pid,
                 num_ports=alsa.snd_seq_client_info_get_num_ports(info),
                 event_lost=alsa.snd_seq_client_info_get_event_lost(info),
+                event_filter=event_filter,
                 )
 
     def _to_alsa(self) -> _snd_seq_client_info_t:
@@ -139,9 +145,205 @@ class ClientInfo:
         alsa.snd_seq_client_info_set_name(info, self.name.encode())
         alsa.snd_seq_client_info_set_broadcast_filter(info, 1 if self.broadcast_filter else 0)
         alsa.snd_seq_client_info_set_error_bounce(info, 1 if self.error_bounce else 0)
+        if self.event_filter:
+            for e_type in self.event_filter:
+                alsa.snd_seq_client_info_event_filter_add(info, e_type)
+        else:
+            alsa.snd_seq_client_info_event_filter_clear(info)
         return info
 
 
+_snd_seq_system_info_t = NewType("_snd_seq_system_info_t", object)
+
+
+@dataclass
+class SystemInfo:
+    """System information.
+
+    Represents data from :alsa:`snd_seq_system_info_t`
+
+    :ivar queues: maximum number of clients
+    :ivar clients: maximum number of ports
+    :ivar ports: maximum number of channels
+    :ivar channels: maximum number of channels
+    :ivar cur_clients: current number of clients
+    :ivar cur_queues: current number of queues
+    """
+    queues: int
+    clients: int
+    ports: int
+    channels: int
+    cur_clients: int
+    cur_queues: int
+
+    @classmethod
+    def _from_alsa(cls, info: _snd_seq_system_info_t):
+        """Create a ClientInfo object from ALSA :alsa:`snd_seq_system_info_t`."""
+        return cls(
+                queues=alsa.snd_seq_system_info_get_queues(info),
+                clients=alsa.snd_seq_system_info_get_clients(info),
+                ports=alsa.snd_seq_system_info_get_ports(info),
+                channels=alsa.snd_seq_system_info_get_channels(info),
+                cur_clients=alsa.snd_seq_system_info_get_cur_clients(info),
+                cur_queues=alsa.snd_seq_system_info_get_cur_queues(info),
+                )
+
+
+_snd_seq_query_subscribe_t = NewType("_snd_seq_query_subscribe_t", object)
+
+
+class SubscriptionQueryType(IntEnum):
+    READ = alsa.SND_SEQ_QUERY_SUBS_READ
+    WRITE = alsa.SND_SEQ_QUERY_SUBS_WRITE
+
+
+@dataclass
+class SubscriptionQuery:
+    """Port subscription (connection) information.
+
+    Represents data from :alsa:`snd_seq_query_subscribe_t`
+
+    :param root: address of the port queried
+    :param type: either :data:`SubscriptionQueryType.READ` or :data:`SubscriptionQueryType.WRITE`
+    :param index: subscription index inside query result
+    :param num_subs: number of subscription in query result
+    :param addr: address of the subscriber
+    :param queue_id: queue id
+    :param exclusive: exclusive access
+    :param time_update: time update enabled
+    :param time_real: user real time stamps
+
+    :ivar root: address of the port queried
+    :ivar type: either :data:`SubscriptionQueryType.READ` or :data:`SubscriptionQueryType.WRITE`
+    :ivar index: subscription index inside query result
+    :ivar num_subs: number of subscription in query result
+    :ivar addr: address of the subscriber
+    :ivar queue_id: queue id
+    :ivar exclusive: exclusive access
+    :ivar time_update: time update enabled
+    :ivar time_real: user real time stamps
+    """
+    root: Address = Address(0, 0)
+    type: SubscriptionQueryType = SubscriptionQueryType.READ
+    index: int = 0
+    num_subs: int = 0
+    addr: Address = Address(0, 0)
+    queue_id: int = 0
+    exclusive: bool = False
+    time_update: bool = False
+    time_real: bool = False
+
+    def __post_init__(self):
+        self.root = Address(self.root)
+        self.addr = Address(self.addr)
+        self.exlusive = bool(self.exclusive)
+        self.time_update = bool(self.time_update)
+        self.time_real = bool(self.time_real)
+
+    @classmethod
+    def _from_alsa(cls, query: _snd_seq_query_subscribe_t):
+        """Create a ClientInfo object from ALSA :alsa:`snd_seq_query_subscribe_t`."""
+        a_root = alsa.snd_seq_query_subscribe_get_root(query)
+        a_addr = alsa.snd_seq_query_subscribe_get_addr(query)
+        return cls(
+                root=Address(a_root.client, a_root.port),
+                type=SubscriptionQueryType(alsa.snd_seq_query_subscribe_get_type(query)),
+                index=alsa.snd_seq_query_subscribe_get_index(query),
+                num_subs=alsa.snd_seq_query_subscribe_get_num_subs(query),
+                addr=Address(a_addr.client, a_addr.port),
+                queue_id=alsa.snd_seq_query_subscribe_get_queue(query),
+                exclusive=alsa.snd_seq_query_subscribe_get_exclusive(query),
+                time_update=alsa.snd_seq_query_subscribe_get_time_update(query),
+                time_real=alsa.snd_seq_query_subscribe_get_time_real(query),
+                )
+
+    def _to_alsa(self) -> _snd_seq_query_subscribe_t:
+        """Create a an ALSA :alsa:`snd_seq_query_subscribe_t` object from self."""
+        query_p = ffi.new("snd_seq_query_subscribe_t **")
+        err = alsa.snd_seq_query_subscribe_malloc(query_p)
+        _check_alsa_error(err)
+        query = ffi.gc(query_p[0], alsa.snd_seq_query_subscribe_free)
+        alsa.snd_seq_query_subscribe_set_client(query, self.root.client_id)
+        alsa.snd_seq_query_subscribe_set_port(query, self.root.port_id)
+        alsa.snd_seq_query_subscribe_set_type(query, self.type)
+        alsa.snd_seq_query_subscribe_set_index(query, self.index)
+        return query
+
+
+_snd_seq_client_pool_t = NewType("_snd_seq_client_pool_t", object)
+
+
+@dataclass
+class ClientPool:
+    """Client kernel-side memory pool information.
+
+    Represents data from :alsa:`snd_seq_client_pool_t`
+
+    :ivar client_id: client id
+    :ivar output_pool: output pool size
+    :ivar input_pool: input pool size
+    :ivar output_room: output pool room size
+    :ivar output_free: amount of free space in the output pool
+    :ivar input_free: amount of free space in the input pool
+    """
+    client_id: int = 0
+    output_pool: int = 0
+    input_pool: int = 0
+    output_room: int = 0
+    output_free: int = 0
+    input_free: int = 0
+
+    @classmethod
+    def _from_alsa(cls, pool: _snd_seq_client_pool_t):
+        """Create a ClientInfo object from ALSA :alsa:`snd_seq_client_pool_t`."""
+        return cls(
+                client_id=alsa.snd_seq_client_pool_get_client(pool),
+                output_pool=alsa.snd_seq_client_pool_get_output_pool(pool),
+                input_pool=alsa.snd_seq_client_pool_get_input_pool(pool),
+                output_room=alsa.snd_seq_client_pool_get_output_room(pool),
+                output_free=alsa.snd_seq_client_pool_get_output_free(pool),
+                input_free=alsa.snd_seq_client_pool_get_input_free(pool),
+                )
+
+    def _to_alsa(self) -> _snd_seq_client_pool_t:
+        """Create a an ALSA :alsa:`snd_seq_client_pool_t` object from self."""
+        pool_p = ffi.new("snd_seq_client_pool_t **")
+        err = alsa.snd_seq_client_pool_malloc(pool_p)
+        _check_alsa_error(err)
+        pool = ffi.gc(pool_p[0], alsa.snd_seq_client_pool_free)
+        alsa.snd_seq_client_pool_set_output_pool(pool, self.output_pool)
+        alsa.snd_seq_client_pool_set_input_pool(pool, self.input_pool)
+        alsa.snd_seq_client_pool_set_output_room(pool, self.output_room)
+        return pool
+
+
+class RemoveCondition(IntFlag):
+    INPUT = alsa.SND_SEQ_REMOVE_INPUT
+    OUTPUT = alsa.SND_SEQ_REMOVE_OUTPUT
+    DEST = alsa.SND_SEQ_REMOVE_DEST
+    DEST_CHANNEL = alsa.SND_SEQ_REMOVE_DEST_CHANNEL
+    TIME_BEFORE = alsa.SND_SEQ_REMOVE_TIME_BEFORE
+    TIME_AFTER = alsa.SND_SEQ_REMOVE_TIME_AFTER
+    TIME_TICK = alsa.SND_SEQ_REMOVE_TIME_TICK
+    EVENT_TYPE = alsa.SND_SEQ_REMOVE_EVENT_TYPE
+    IGNORE_OFF = alsa.SND_SEQ_REMOVE_IGNORE_OFF
+    TAG_MATCH = alsa.SND_SEQ_REMOVE_TAG_MATCH
+
+
+@dataclass
+class RemoveEvents:
+    """Events removal condition.
+
+    Represents :alsa:`snd_seq_remove_events_t` data."""
+    condition: RemoveCondition = RemoveCondition(0)
+    queue_id: int = 0
+    time: Union[RealTime, int] = RealTime(0, 0)
+    dest: AddressType = Address(0, 0)
+    channel: int = 0
+    event_type: EventType = EventType.NONE
+    tag: int = 0
+
+
 class SequencerClientBase:
     """Base class for :class:`SequencerClient` and :class:`AsyncSequencerClient`.
 
@@ -159,6 +361,7 @@ class SequencerClientBase:
     _handle_p: _snd_seq_t_p
     _fd: int = -1
     _event_parser: Optional[_snd_midi_event_t] = None
+    _queues: MutableMapping[int, Queue]
 
     def __init__(
             self,
@@ -166,7 +369,7 @@ class SequencerClientBase:
             streams: int = StreamOpenType.DUPLEX,
             mode: int = OpenMode.NONBLOCK,
             sequencer_name: str = "default"):
-
+        self._queues = WeakValueDictionary()
         client_name_b = client_name.encode("utf-8")
         sequencer_name_b = sequencer_name.encode("utf-8")
         self._handle_p = ffi.new("snd_seq_t **", ffi.NULL)
@@ -224,9 +427,60 @@ class SequencerClientBase:
             err = alsa.snd_midi_event_new(1024, parser_p)
             _check_alsa_error(err)
             parser = ffi.gc(parser_p[0], alsa.snd_midi_event_free)
+            alsa.snd_midi_event_no_status(parser, 1)
             self._event_parser = parser
         return parser
 
+    def get_sequencer_name(self) -> str:
+        """Get sequencer name.
+
+        Wraps :alsa:`snd_seq_name`."""
+        self._check_handle()
+        return ffi.string(alsa.snd_seq_name(self.handle)).decode()
+
+    def get_sequencer_type(self) -> SequencerType:
+        """Get sequencer type.
+
+        Wraps :alsa:`snd_seq_type`."""
+        self._check_handle()
+        result = alsa.snd_seq_type(self.handle)
+        _check_alsa_error(result)
+        return SequencerType(result)
+
+    def get_output_buffer_size(self) -> int:
+        """Get output buffer size.
+
+        Wraps :alsa:`snd_seq_get_output_buffer_size`."""
+        self._check_handle()
+        return alsa.snd_seq_get_output_buffer_size(self.handle)
+
+    def set_output_buffer_size(self, size: int):
+        """Change output buffer size.
+
+        :param size: the size of output buffer in bytes
+
+        Wraps :alsa:`snd_seq_set_output_buffer_size`."""
+        self._check_handle()
+        err = alsa.snd_seq_set_output_buffer_size(self.handle, size)
+        _check_alsa_error(err)
+
+    def get_input_buffer_size(self) -> int:
+        """Get input buffer size.
+
+        Wraps :alsa:`snd_seq_get_input_buffer_size`."""
+        self._check_handle()
+        return alsa.snd_seq_get_input_buffer_size(self.handle)
+
+    def set_input_buffer_size(self, size: int):
+        """Change input buffer size.
+
+        :param size: the size of input buffer in bytes
+
+        Wraps :alsa:`snd_seq_set_input_buffer_size`."""
+        self._check_handle()
+        err = alsa.snd_seq_set_input_buffer_size(self.handle, size)
+        _check_alsa_error(err)
+
     def create_port(self,
                     name: str,
                     caps: PortCaps = RW_PORT,
@@ -297,22 +551,59 @@ class SequencerClientBase:
         _check_alsa_error(port)
         return Port(self, port)
 
-    def create_queue(self, name: str = None) -> Queue:
+    def create_queue(self, name: str = None, info: QueueInfo = None) -> Queue:
         """Create a new queue.
 
-        Wraps :alsa:`snd_seq_alloc_named_queue` or :alsa:`snd_seq_alloc_queue`.
+        Wraps :alsa:`snd_seq_create_queue`, :alsa:`snd_seq_alloc_named_queue`
+        or :alsa:`snd_seq_alloc_queue`.
 
         :param name: queue name
+        :param info: queue creation parameters
 
         :return: queue object created.
         """
         self._check_handle()
-        if name is not None:
+        if info is not None:
+            if name is not None:
+                raise ValueError("'name' and 'info' cannot be both provided")
+            alsa_info = info._to_alsa(client_id=self.client_id)
+            queue = alsa.snd_seq_create_queue(self.handle, alsa_info)
+        elif name is not None:
             queue = alsa.snd_seq_alloc_named_queue(self.handle, name.encode("utf-8"))
         else:
             queue = alsa.snd_seq_alloc_queue(self.handle)
         _check_alsa_error(queue)
-        return Queue(self, queue)
+        assert queue not in self._queues
+        result = Queue(self, queue, _own=True)
+        self._queues[queue] = result
+        return result
+
+    def get_queue(self, queue_id: int) -> Queue:
+        """Get a queue object by id.
+
+        Wraps :alsa:`snd_seq_set_queue_usage` when a new object is returned.
+
+        :return: queue with a provided name
+        """
+        self._check_handle()
+        queue = self._queues.get(queue_id)
+        if queue is None:
+            err = alsa.snd_seq_set_queue_usage(self.handle, queue_id, 1)
+            _check_alsa_error(err)
+            queue = Queue(self, queue_id, _own=False)
+        return queue
+
+    def get_named_queue(self, name: str) -> Queue:
+        """Get a queue object by name.
+
+        Wraps :alsa:`snd_seq_query_named_queue` and :alsa:`snd_seq_set_queue_usage`.
+
+        :param name: queue name
+
+        :return: queue by the provided name
+        """
+        queue_id = self.query_named_queue(name)
+        return self.get_queue(queue_id)
 
     def drop_input(self):
         """Remove all incoming events in the input buffer and sequencer queue.
@@ -342,7 +633,7 @@ class SequencerClientBase:
         _check_alsa_error(err)
 
     def drop_output(self):
-        """Remove all events from the output buffer.
+        """Remove all events from the output buffer (client and kernel side).
 
         Wraps :alsa:`snd_seq_drop_output`.
         """
@@ -350,6 +641,15 @@ class SequencerClientBase:
         err = alsa.snd_seq_drop_output(self.handle)
         _check_alsa_error(err)
 
+    def drop_output_buffer(self):
+        """Remove all events from the output buffer (client side only).
+
+        Wraps :alsa:`snd_seq_drop_output_buffer`.
+        """
+        self._check_handle()
+        err = alsa.snd_seq_drop_output_buffer(self.handle)
+        _check_alsa_error(err)
+
     def _event_input(self, prefer_bytes: bool = False) -> Tuple[int, Optional[Event]]:
         buf = ffi.new("snd_seq_event_t**", ffi.NULL)
         result = alsa.snd_seq_event_input(self.handle, buf)
@@ -394,6 +694,26 @@ class SequencerClientBase:
         _check_alsa_error(result)
         return event
 
+    def event_input_pending(self, fetch_sequencer: bool = False) -> int:
+        """Check events in input buffer.
+
+        If events remain on the input buffer of user-space, this method returns
+        the total byte size of events on it. If fetch_sequencer argument is
+        :data:`True`, this method checks the presence of events on sequencer FIFO
+        When events exist, they are transferred to the input buffer, and the
+        number of received events are returned. If fetch_sequencer argument is
+        zero and no events remain on the input buffer, function simply returns
+        zero.
+
+        Wraps :alsa:`snd_seq_event_input_pending`.
+
+        :param fetch_sequencer: check events pending on the kernel side too
+        """
+        self._check_handle()
+        result = alsa.snd_seq_event_input_pending(self.handle, int(fetch_sequencer))
+        _check_alsa_error(result)
+        return result
+
     def _prepare_event(self,
                        event: Event,
                        queue: Union['Queue', int] = None,
@@ -574,6 +894,90 @@ class SequencerClientBase:
                 break
         return result
 
+    def event_output_pending(self) -> int:
+        """Return the size of pending events on output buffer.
+
+        Wraps :alsa:`snd_seq_event_output_pending`."""
+        self._check_handle()
+        result = alsa.snd_seq_event_output_pending(self.handle)
+        _check_alsa_error(result)
+        return result
+
+    def extract_output(self) -> Event:
+        """Extract the first event in output buffer.
+
+        Wraps :alsa:`snd_seq_extract_output`."""
+        self._check_handle()
+        buf = ffi.new("snd_seq_event_t**", ffi.NULL)
+        result = alsa.snd_seq_extract_output(self.handle, buf)
+        _check_alsa_error(result)
+        alsa_event = buf[0]
+        try:
+            cls = Event._specialized.get(buf[0].type, Event)
+            return cls._from_alsa(alsa_event)
+        finally:
+            alsa.snd_seq_free_event(alsa_event)
+
+    def get_system_info(self) -> SystemInfo:
+        """Obtain information about the sequencer.
+
+        Wraps :alsa:`snd_seq_system_info`.
+
+        :return: system information
+        """
+        info_p = ffi.new("snd_seq_system_info_t **")
+        err = alsa.snd_seq_system_info_malloc(info_p)
+        _check_alsa_error(err)
+        info = ffi.gc(info_p[0], alsa.snd_seq_system_info_free)
+        err = alsa.snd_seq_system_info(self.handle, info)
+        _check_alsa_error(err)
+        result = SystemInfo._from_alsa(info)
+        return result
+
+    def get_client_info(self, client_id: Optional[int] = None) -> ClientInfo:
+        """Obtain information about a client.
+
+        Wraps :alsa:`snd_seq_get_client_info` or :alsa:`snd_seq_get_any_client_info`.
+
+        :param client_id: client to get information about. Default: self.
+
+        :return: client information
+        """
+        info_p = ffi.new("snd_seq_client_info_t **")
+        err = alsa.snd_seq_client_info_malloc(info_p)
+        _check_alsa_error(err)
+        info = ffi.gc(info_p[0], alsa.snd_seq_client_info_free)
+        if client_id is None:
+            err = alsa.snd_seq_get_client_info(self.handle, info)
+        else:
+            err = alsa.snd_seq_get_any_client_info(self.handle, client_id, info)
+        _check_alsa_error(err)
+        result = ClientInfo._from_alsa(info)
+        return result
+
+    def set_client_info(self, info: ClientInfo):
+        """Set client information including event filter.
+
+        Wraps :alsa:`snd_seq_set_client_info`.
+
+        :param info: new client info
+        """
+        self._check_handle()
+        a_info = info._to_alsa()
+        err = alsa.snd_seq_set_client_info(self.handle, a_info)
+        _check_alsa_error(err)
+
+    def set_client_event_filter(self, event_type: EventType):
+        """Add an event to client's event filter.
+
+        Wraps :alsa:`snd_seq_set_client_event_filter`.
+
+        :param event_type: event type to accept
+        """
+        self._check_handle()
+        err = alsa.snd_seq_set_client_event_filter(self.handle, int(event_type))
+        _check_alsa_error(err)
+
     @overload
     def query_next_client(self, previous: ClientInfo) -> Optional[ClientInfo]:
         ...
@@ -613,6 +1017,8 @@ class SequencerClientBase:
 
         Wraps :alsa:`snd_seq_get_port_info` or :alsa:`snd_seq_get_any_port_info`.
 
+        :param port: port to get information about
+
         :return: port information
         """
         if isinstance(port, int):
@@ -636,8 +1042,6 @@ class SequencerClientBase:
         """Set information about a specific own port.
 
         Wraps :alsa:`snd_seq_set_port_info`.
-
-        :return: port information
         """
         if isinstance(port, int):
             port_id = port
@@ -883,6 +1287,313 @@ class SequencerClientBase:
         return self._subunsub_port(alsa.snd_seq_unsubscribe_port,
                                    sender, dest)
 
+    def query_port_subscribers(self,
+                               query: SubscriptionQuery
+                               ) -> SubscriptionQuery:
+        """Queries the subscribers subscribers to a port.
+
+        At least, the client id, the port id, the index number and the query
+        type must be set in to perform a proper query. As the query type,
+        :data:`SubscriptionQueryType.READ` or
+        :data:`SubscriptionQueryType.WRITE` can be specified to check whether
+        the readers or the writers to the port. To query the first
+        subscription, set 0 to the index number. To list up all the
+        subscriptions, call this method with the index numbers from 0 until
+        this raises :class:`ALSAError`. Or use :meth:`list_port_subscribers()`
+        instead.
+
+        Wraps :alsa:`snd_seq_query_port_subscribers`.
+
+        :param query: Query parameters
+        """
+        self._check_handle()
+        a_query = query._to_alsa()
+        err = alsa.snd_seq_query_port_subscribers(self.handle, a_query)
+        _check_alsa_error(err)
+        result = SubscriptionQuery._from_alsa(a_query)
+        return result
+
+    def list_port_subscribers(self,
+                              port: AddressType,
+                              type: Optional[SubscriptionQueryType] = None,
+                              ) -> List[SubscriptionQuery]:
+        """Lists subscribers accessing a port.
+
+        Wraps :alsa:`snd_seq_query_port_subscribers`.
+
+        :param port: Port address to query
+        :param type: limit query to the specific type
+        """
+        if type is None:
+            types = [SubscriptionQueryType.READ, SubscriptionQueryType.WRITE]
+        else:
+            types = [type]
+
+        self._check_handle()
+        query_p = ffi.new("snd_seq_query_subscribe_t **")
+        err = alsa.snd_seq_query_subscribe_malloc(query_p)
+        _check_alsa_error(err)
+        query = ffi.gc(query_p[0], alsa.snd_seq_query_subscribe_free)
+
+        client_id, port_id = Address(port)
+
+        result = []
+
+        for type in types:
+            alsa.snd_seq_query_subscribe_set_client(query, client_id)
+            alsa.snd_seq_query_subscribe_set_port(query, port_id)
+            alsa.snd_seq_query_subscribe_set_type(query, type)
+            i = 0
+            while True:
+                alsa.snd_seq_query_subscribe_set_index(query, i)
+                err = alsa.snd_seq_query_port_subscribers(self.handle, query)
+                if err < 0:
+                    break
+                result.append(SubscriptionQuery._from_alsa(query))
+                i += 1
+        return result
+
+    def get_client_pool(self) -> ClientPool:
+        """Obtain the pool information of the client.
+
+        Wraps :alsa:`snd_seq_get_client_pool`.
+        """
+        self._check_handle()
+        pool_p = ffi.new("snd_seq_client_pool_t **")
+        err = alsa.snd_seq_client_pool_malloc(pool_p)
+        _check_alsa_error(err)
+        a_pool = ffi.gc(pool_p[0], alsa.snd_seq_client_pool_free)
+        err = alsa.snd_seq_get_client_pool(self.handle, a_pool)
+        _check_alsa_error(err)
+        return ClientPool._from_alsa(a_pool)
+
+    def set_client_pool(self, pool: ClientPool):
+        """Change pool settings of the client.
+
+        :param pool: new pool settings
+
+        Wraps :alsa:`snd_seq_set_client_pool`.
+        """
+        self._check_handle()
+        a_pool = pool._to_alsa()
+        err = alsa.snd_seq_set_client_pool(self.handle, a_pool)
+        _check_alsa_error(err)
+
+    def set_client_pool_output(self, size: int):
+        """Change output pool size for the client.
+
+        :param size: requested pool size
+
+        Wraps :alsa:`snd_seq_set_client_pool_output`
+        """
+        err = alsa.snd_seq_set_client_pool_output(self.handle, size)
+        _check_alsa_error(err)
+
+    def set_client_pool_output_room(self, size: int):
+        """Change output pool room size for the client.
+
+        :param size: requested room size
+
+        Wraps :alsa:`snd_seq_set_client_pool_output_room`
+        """
+        err = alsa.snd_seq_set_client_pool_output_room(self.handle, size)
+        _check_alsa_error(err)
+
+    def set_client_pool_input(self, size: int):
+        """Change input pool size for the client.
+
+        :param size: requested pool size
+
+        Wraps :alsa:`snd_seq_set_client_pool_input`
+        """
+        err = alsa.snd_seq_set_client_pool_input(self.handle, size)
+        _check_alsa_error(err)
+
+    def get_queue_info(self, queue_id: int) -> QueueInfo:
+        """Obtain queue attributes.
+
+        Wraps :alsa:`snd_seq_get_queue_info`.
+
+        :param queue_id: identifier of the queue
+
+        :return: queue information
+        """
+        self._check_handle()
+        info_p = ffi.new("snd_seq_queue_info_t **")
+        err = alsa.snd_seq_queue_info_malloc(info_p)
+        _check_alsa_error(err)
+        info = ffi.gc(info_p[0], alsa.snd_seq_queue_info_free)
+        err = alsa.snd_seq_get_queue_info(self.handle, queue_id, info)
+        _check_alsa_error(err)
+        result = QueueInfo._from_alsa(info)
+        return result
+
+    def set_queue_info(self, queue_id: int, info: QueueInfo):
+        """Change queue attributes.
+
+        :param queue_id: identifier of the queue
+
+        Wraps :alsa:`snd_seq_set_queue_info`.
+        """
+        self._check_handle()
+        alsa_info = info._to_alsa(client_id=self.client_id)
+        err = alsa.snd_seq_set_queue_info(self.handle, queue_id, alsa_info)
+        _check_alsa_error(err)
+
+    def query_named_queue(self, name: str) -> int:
+        """Query queue by name.
+
+        Wraps :alsa:`snd_seq_query_named_queue`.
+
+        :param name: queue name
+
+        :return: queue id
+        """
+        self._check_handle()
+        queue_id = alsa.snd_seq_query_named_queue(self.handle, name.encode("utf-8"))
+        _check_alsa_error(queue_id)
+        return queue_id
+
+    def get_queue_status(self, queue_id: int) -> QueueStatus:
+        """Get queue status.
+
+        Wraps :alsa:`snd_seq_get_queue_status`.
+
+        :param queue_id: identifier of the queue
+        """
+        self._check_handle()
+        status_p = ffi.new("snd_seq_queue_status_t **")
+        err = alsa.snd_seq_queue_status_malloc(status_p)
+        _check_alsa_error(err)
+        status = ffi.gc(status_p[0], alsa.snd_seq_queue_status_free)
+        err = alsa.snd_seq_get_queue_status(self.handle, queue_id, status)
+        _check_alsa_error(err)
+        result = QueueStatus._from_alsa(status)
+        return result
+
+    @overload
+    def remove_events(self, condition: RemoveEvents):
+        ...
+
+    @overload
+    def remove_events(self,
+                      condition: RemoveCondition = None,
+                      queue: Union[Queue, int] = None,
+                      time: RealTime = None,
+                      dest: AddressType = None,
+                      channel: int = None,
+                      event_type: EventType = None,
+                      tag: int = None,
+                      before: bool = False,
+                      ignore_off: bool = None,
+                      ):
+        ...
+
+    def remove_events(self,
+                      condition: Union[RemoveEvents, RemoveCondition] = None,
+                      queue: Union[Queue, int] = None,
+                      time: Union[RealTime, int] = None,
+                      dest: AddressType = None,
+                      channel: int = None,
+                      event_type: EventType = None,
+                      tag: int = None,
+                      before: bool = False,
+                      ignore_off: bool = None,
+                      ):
+        """Remove selected from output and/or input buffers.
+
+        Wraps :alsa:`snd_seq_remove_events`.
+
+        Note: while this wraps complete ALSA 'remove events' API a lot of
+        possible conditions are not actually implemented in ALSA, especially on
+        the kernel side.
+
+        :param condition: Condition flags. Additional flags will be added, as
+                          needed when other arguments are provided. Can also be
+                          a RemoveEvents objects, to describe whole ALSA
+                          condition.
+        :param queue: Queue (object or id). Seems to be ignored by ALSA.
+        :param time: Time (real or tick) to compare event timestamps to.
+        :param dest: Destination address
+        :param channel: MIDI channel number
+        :param event_type: type of events to remove
+        :param tag: event tag to matching
+        :param before: match events before specified time
+        :param ignore_off: do not remove Note Off events
+        """
+
+        if (isinstance(condition, RemoveEvents)
+                and any((queue, time, dest, channel, event_type, tag))):
+            raise TypeError("condition must not be RemoveEvents object when using other arguments")
+        self._check_handle()
+        cond_p = ffi.new("snd_seq_remove_events_t **")
+        err = alsa.snd_seq_remove_events_malloc(cond_p)
+        _check_alsa_error(err)
+        a_cond = ffi.gc(cond_p[0], alsa.snd_seq_remove_events_free)
+        if isinstance(condition, RemoveEvents):
+            cond_bits = condition.condition
+            queue_id = condition.queue_id
+            time = condition.time
+            dest = condition.dest
+            channel = condition.channel
+            event_type = condition.event_type
+            tag = condition.tag
+        else:
+            if condition is None:
+                cond_bits = RemoveCondition(0)
+            else:
+                cond_bits = condition
+            if dest is not None:
+                cond_bits |= RemoveCondition.DEST
+            if channel is not None:
+                cond_bits |= RemoveCondition.DEST_CHANNEL
+            if time is not None:
+                if before:
+                    cond_bits |= RemoveCondition.TIME_BEFORE
+                else:
+                    cond_bits |= RemoveCondition.TIME_AFTER
+                if isinstance(time, int):
+                    cond_bits |= RemoveCondition.TIME_TICK
+            if event_type is not None:
+                cond_bits |= RemoveCondition.EVENT_TYPE
+            if tag is not None:
+                cond_bits |= RemoveCondition.TAG_MATCH
+            if ignore_off:
+                cond_bits |= RemoveCondition.IGNORE_OFF
+            if isinstance(queue, int):
+                queue_id = queue
+            elif queue is not None:
+                queue_id = queue.queue_id
+            else:
+                queue_id = None
+
+        alsa.snd_seq_remove_events_set_condition(a_cond, cond_bits)
+        if queue_id is not None:
+            alsa.snd_seq_remove_events_set_queue(a_cond, queue_id)
+        if time is not None:
+            a_time = ffi.new("snd_seq_timestamp_t *")
+            if isinstance(time, RealTime):
+                a_time.time.tv_sec = time.seconds
+                a_time.time.tv_nsec = time.nanoseconds
+            else:
+                a_time.tick = time
+            alsa.snd_seq_remove_events_set_time(a_cond, a_time)
+        if dest is not None:
+            dest = Address(dest)
+            a_dest = ffi.new("snd_seq_addr_t *")
+            a_dest.client = dest.client_id
+            a_dest.port = dest.port_id
+            alsa.snd_seq_remove_events_set_dest(a_cond, a_dest)
+        if channel is not None:
+            alsa.snd_seq_remove_events_set_channel(a_cond, channel)
+        if event_type is not None:
+            alsa.snd_seq_remove_events_set_event_type(a_cond, event_type)
+        if tag is not None:
+            alsa.snd_seq_remove_events_set_tag(a_cond, tag)
+        result = alsa.snd_seq_remove_events(self.handle, a_cond)
+        _check_alsa_error(result)
+        return result
+
 
 class SequencerClient(SequencerClientBase):
     """ALSA sequencer client connection.
@@ -1225,4 +1936,6 @@ class AsyncSequencerClient(SequencerClientBase):
         return await self._event_output_wait(func)
 
 
-__all__ = ["SequencerClientBase", "SequencerClient", "ClientInfo", "ClientType"]
+__all__ = ["SequencerClientBase", "SequencerClient", "ClientInfo", "ClientType", "SequencerType",
+           "SystemInfo", "SubscriptionQueryType", "SubscriptionQuery", "ClientPool",
+           "RemoveEvents", "RemoveCondition"]
