@@ -1,10 +1,11 @@
 
 import errno
+import sys
 
 import pytest
 
-from alsa_midi import (ALSAError, AsyncSequencerClient, ClientInfo, ClientType, SequencerClient,
-                       StateError, alsa, ffi)
+from alsa_midi import (ALSAError, AsyncSequencerClient, ClientInfo, ClientPool, ClientType,
+                       SequencerClient, SequencerType, StateError, SystemInfo, alsa, ffi)
 from alsa_midi.client import SequencerClientBase
 from alsa_midi.event import EventType, MidiBytesEvent, NoteOffEvent, NoteOnEvent
 
@@ -63,6 +64,52 @@ def test_client_open_del_alsa(alsa_seq_state):
     assert client_id not in alsa_seq_state.clients
 
 
+@pytest.mark.require_alsa_seq
+def test_sequencer_name():
+    client = SequencerClient("test123")
+    assert client.get_sequencer_name() == "default"
+    client.close()
+
+
+@pytest.mark.require_alsa_seq
+def test_sequencer_type():
+    client = SequencerClient("test123")
+    assert client.get_sequencer_type() == SequencerType.HW
+    client.close()
+
+
+@pytest.mark.require_alsa_seq
+def test_output_bufer_size():
+    client = SequencerClient("test")
+    ev_size = ffi.sizeof("snd_seq_event_t")
+    client.set_output_buffer_size(100 * ev_size)
+    assert client.get_output_buffer_size() == 100 * ev_size
+    client.set_output_buffer_size(10 * ev_size)
+    assert client.get_output_buffer_size() == 10 * ev_size
+    with pytest.raises(OverflowError):
+        client.set_output_buffer_size(-1)
+    with pytest.raises(ALSAError) as err:
+        client.set_output_buffer_size(sys.maxsize)
+    assert err.value.errnum == -errno.ENOMEM
+    client.close()
+
+
+@pytest.mark.require_alsa_seq
+def test_input_bufer_size():
+    client = SequencerClient("test")
+    ev_size = ffi.sizeof("snd_seq_event_t")
+    client.set_input_buffer_size(100 * ev_size)
+    assert client.get_input_buffer_size() == 100 * ev_size
+    client.set_input_buffer_size(10 * ev_size)
+    assert client.get_input_buffer_size() == 10 * ev_size
+    with pytest.raises(OverflowError):
+        client.set_input_buffer_size(-1)
+    with pytest.raises(ALSAError) as err:
+        client.set_input_buffer_size(sys.maxsize)
+    assert err.value.errnum == -errno.ENOMEM
+    client.close()
+
+
 @pytest.mark.require_alsa_seq
 def test_open_blocking():
     with pytest.raises(ValueError):
@@ -229,6 +276,7 @@ def test_client_info():
     assert info.pid is None
     assert info.num_ports == 0
     assert info.event_lost == 0
+    assert info.event_filter is None
 
     # test initializing all attributes
     info = ClientInfo(client_id=15,
@@ -239,7 +287,8 @@ def test_client_info():
                       card_id=8,
                       pid=100,
                       num_ports=5,
-                      event_lost=7)
+                      event_lost=7,
+                      event_filter={EventType.NOTEON, EventType.NOTEOFF})
 
     assert info.client_id == 15
     assert info.name == "client_info_test2"
@@ -250,6 +299,7 @@ def test_client_info():
     assert info.pid == 100
     assert info.num_ports == 5
     assert info.event_lost == 7
+    assert info.event_filter == {EventType.NOTEON, EventType.NOTEOFF}
 
     # test _to_alsa (only some values are writable to the ALSA struct)
     info = ClientInfo(client_id=17,
@@ -267,6 +317,19 @@ def test_client_info():
     assert ffi.string(alsa.snd_seq_client_info_get_name(alsa_info)) == b"client_info_test3"
     assert alsa.snd_seq_client_info_get_broadcast_filter(alsa_info) == 1
     assert alsa.snd_seq_client_info_get_error_bounce(alsa_info) == 0
+    assert alsa.snd_seq_client_info_get_event_filter(alsa_info) == ffi.NULL
+
+    # test _to_alsa with event filter
+    info = ClientInfo(client_id=17,
+                      name="xxx",
+                      event_filter={EventType.NOTEON, EventType.NOTEOFF})
+    assert info.event_filter == {EventType.NOTEON, EventType.NOTEOFF}
+
+    alsa_info = info._to_alsa()
+    assert alsa.snd_seq_client_info_get_event_filter(alsa_info) != ffi.NULL
+    assert alsa.snd_seq_client_info_event_filter_check(alsa_info, EventType.NOTEON)
+    assert alsa.snd_seq_client_info_event_filter_check(alsa_info, EventType.NOTEOFF)
+    assert not alsa.snd_seq_client_info_event_filter_check(alsa_info, EventType.CONTROLLER)
 
     # test _from_alsa (only the attributes we can set)
     info_p = ffi.new("snd_seq_client_info_t **")
@@ -284,6 +347,91 @@ def test_client_info():
     assert info.broadcast_filter is True
     assert info.error_bounce is True
     assert info.type == ClientType._UNSET
+    assert info.event_filter is None
+
+    # test _from_alsa with filter
+    info_p = ffi.new("snd_seq_client_info_t **")
+    err = alsa.snd_seq_client_info_malloc(info_p)
+    assert err >= 0
+    alsa_info = ffi.gc(info_p[0], alsa.snd_seq_client_info_free)
+    alsa.snd_seq_client_info_event_filter_add(alsa_info, EventType.NOTEON)
+    alsa.snd_seq_client_info_event_filter_add(alsa_info, EventType.CONTROLLER)
+    info = ClientInfo._from_alsa(alsa_info)
+
+    assert info.event_filter == {EventType.NOTEON, EventType.CONTROLLER}
+
+
+@pytest.mark.require_alsa_seq
+def test_get_client_info(alsa_seq_state):
+    client = SequencerClient("test")
+    alsa_seq_state.load()
+
+    info = client.get_client_info()
+    assert isinstance(info, ClientInfo)
+    assert info.name == "test"
+
+    for client_id, alsa_client in alsa_seq_state.clients.items():
+        info = client.get_client_info(client_id)
+        assert isinstance(info, ClientInfo)
+        assert info.name == alsa_client.name
+        assert info.type is not None
+        assert info.type.name.lower() == alsa_client.type.lower()
+
+    client.close()
+
+
+@pytest.mark.require_alsa_seq
+def test_set_client_info(alsa_seq_state):
+    client = SequencerClient("test")
+    alsa_seq_state.load()
+
+    info = ClientInfo(client_id=0,
+                      name="new name",
+                      broadcast_filter=True,
+                      error_bounce=True)
+
+    client.set_client_info(info)
+
+    info = client.get_client_info()
+    assert isinstance(info, ClientInfo)
+    assert info.name == "new name"
+    assert info.broadcast_filter is True
+    assert info.error_bounce is True
+
+    client.close()
+
+
+@pytest.mark.require_alsa_seq
+def test_set_client_event_filter():
+    client = SequencerClient("test")
+
+    info = client.get_client_info()
+    assert info.event_filter is None
+
+    client.set_client_event_filter(EventType.NOTEON)
+    client.set_client_event_filter(EventType.NOTEOFF)
+
+    info = client.get_client_info()
+    assert info.event_filter == {EventType.NOTEON, EventType.NOTEOFF}
+
+    client.close()
+
+
+@pytest.mark.require_alsa_seq
+def test_get_system_info(alsa_seq_state):
+    client = SequencerClient("test")
+    alsa_seq_state.load()
+
+    info = client.get_system_info()
+    assert isinstance(info, SystemInfo)
+    assert isinstance(info.queues, int)
+    assert info.clients == alsa_seq_state.max_clients
+    assert isinstance(info.ports, int)
+    assert isinstance(info.channels, int)
+    assert info.cur_clients == alsa_seq_state.cur_clients
+    assert isinstance(info.cur_queues, int)
+
+    client.close()
 
 
 @pytest.mark.require_alsa_seq
@@ -399,3 +547,68 @@ def test_port_subscribe_unsubscribe(alsa_seq_state):
     finally:
         c2.close()
         c1.close()
+
+
+@pytest.mark.require_alsa_seq
+def test_client_pool():
+    client = SequencerClient("test")
+
+    pool1 = client.get_client_pool()
+    assert isinstance(pool1, ClientPool)
+    assert pool1.client_id == client.client_id
+    assert pool1.output_pool > 0
+    assert pool1.input_pool > 0
+    assert pool1.output_room >= 0 and pool1.output_room <= pool1.output_pool
+    # assert pool1.output_free == pool1.output_pool  # ??
+    assert pool1.input_free > 0 and pool1.input_free <= pool1.input_pool
+
+    pool2 = ClientPool(output_pool=pool1.output_pool * 3,
+                       input_pool=pool1.input_pool * 4,
+                       output_room=pool1.output_pool * 2)
+    client.set_client_pool(pool2)
+
+    pool3 = client.get_client_pool()
+    assert isinstance(pool3, ClientPool)
+    assert pool3.client_id == client.client_id
+    assert pool3.output_pool == pool1.output_pool * 3
+    assert pool3.input_pool == pool1.input_pool * 4
+    assert pool3.output_room == pool1.output_pool * 2
+
+
+@pytest.mark.require_alsa_seq
+def test_set_client_pool_output():
+    client = SequencerClient("test")
+
+    client.set_client_pool_output(100)
+    pool = client.get_client_pool()
+    assert pool.output_pool == 100
+
+    client.set_client_pool_output(200)
+    pool = client.get_client_pool()
+    assert pool.output_pool == 200
+
+
+@pytest.mark.require_alsa_seq
+def test_set_client_pool_output_room():
+    client = SequencerClient("test")
+
+    client.set_client_pool_output_room(100)
+    pool = client.get_client_pool()
+    assert pool.output_room == 100
+
+    client.set_client_pool_output_room(200)
+    pool = client.get_client_pool()
+    assert pool.output_room == 200
+
+
+@pytest.mark.require_alsa_seq
+def test_set_client_pool_input():
+    client = SequencerClient("test")
+
+    client.set_client_pool_input(100)
+    pool = client.get_client_pool()
+    assert pool.input_pool == 100
+
+    client.set_client_pool_input(200)
+    pool = client.get_client_pool()
+    assert pool.input_pool == 200
